import shutil
import subprocess
from collections import OrderedDict
from logging import getLogger
from pathlib import Path

import attr
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from cortexpy.test import runner

from abeona.cli import main as abeona_main
from .expectations import Traversals, Fastas, TraversalsAndFastas

logger = getLogger()


@attr.s(slots=True)
class Mccortex:
    kmer_size = attr.ib(3)
    sequences = attr.ib(attr.Factory(OrderedDict))
    sequence_pairs = attr.ib(attr.Factory(OrderedDict))
    mccortex_bin = attr.ib('mccortex')

    def assert_single_or_paired(self):
        if len(self.sequences) > 0 and len(self.sequence_pairs) > 0:
            raise NotImplementedError(
                'Specifying both single and paired end reads to driver is not implemented')

    @property
    def is_paired(self):
        self.assert_single_or_paired()
        assert len(self.sequence_pairs) != 0 or len(self.sequences) != 0
        return len(self.sequence_pairs) > 0

    def with_kmer_size(self, kmer_size):
        self.kmer_size = kmer_size
        return self

    def with_dna_sequence(self, sequence, *, name='sample_0'):
        if name not in self.sequences:
            self.sequences[name] = []
        self.sequences[name].append([sequence])
        return self

    def with_dna_sequence_pair(self, sequence1, sequence2, name='sample_0'):
        if name not in self.sequence_pairs:
            self.sequence_pairs[name] = []
        self.sequence_pairs[name].append([sequence1, sequence2])
        return self

    def build(self, dir):
        dir = Path(dir)
        dir.mkdir(exist_ok=True)
        self.assert_single_or_paired()
        if len(self.sequence_pairs) == 0:
            sequences = self.sequences
        else:
            sequences = self.sequence_pairs

        mccortex_args = f'build --force --sort --kmer {self.kmer_size}'
        for name, dna_sequence_tuple_list in sequences.items():
            mccortex_args += f' --sample {name}'
            is_paired = len(dna_sequence_tuple_list[0]) == 2
            fa1 = dir / f'input.{name}.1.fasta'
            fa2 = dir / f'input.{name}.2.fasta'
            with open(fa1, 'w') as fh1:
                with open(fa2, 'w') as fh2:
                    for tuple in dna_sequence_tuple_list:
                        fh1.write(SeqRecord(Seq(tuple[0]), id='0', description='').format('fasta'))
                        if is_paired:
                            fh2.write(
                                SeqRecord(Seq(tuple[1]), id='0', description='').format('fasta'))
            if is_paired:
                mccortex_args += f' -2 {fa1}:{fa2}'
            else:
                mccortex_args += f' -1 {fa1}'

        output_graph = str(dir / 'output.ctx')
        mccortex_args += f' {output_graph}'

        ret = runner.Mccortex(self.kmer_size, mccortex_bin=self.mccortex_bin) \
            .run(mccortex_args.split())
        logger.debug('\n' + ret.stdout.decode())
        logger.debug('\n' + ret.stderr.decode())

        ret = runner.Mccortex(self.kmer_size, mccortex_bin=self.mccortex_bin) \
            .view(output_graph)
        logger.debug('\n' + ret.stdout.decode())

        return output_graph


@attr.s(slots=True)
class CleanMccortex:
    mc_builder = attr.ib()
    min_unitig_coverage = attr.ib(0)
    min_tip_length = attr.ib(0)

    @classmethod
    def from_mccortex_builder(cls, builder):
        return cls(mc_builder=builder)

    def __getattr__(self, item):
        return getattr(self.mc_builder, item)

    def with_min_unitig_coverage(self, n):
        self.min_unitig_coverage = n
        return self

    def build(self, dir):
        dir = Path(dir)
        dir.mkdir(exist_ok=True)
        full_graph = self.mc_builder.build(dir)
        clean_graph = dir / 'clean.ctx'

        ret = runner.Mccortex(self.mc_builder.kmer_size, mccortex_bin=self.mc_builder.mccortex_bin) \
            .clean(graphs=[full_graph], out=clean_graph, clip_tips_shorter_than=self.min_tip_length,
                   unitigs_with_mean_coverage_less_than=self.min_unitig_coverage)
        logger.debug('\n' + ret.stdout.decode())
        return clean_graph

@attr.s(slots=True)
class SubgraphTestDriver:
    tmpdir = attr.ib()
    builder = attr.ib(attr.Factory(Mccortex))
    initial_contigs = attr.ib(None)

    def __getattr__(self, item):
        if item in ['with_kmer_size', 'with_dna_sequence']:
            return getattr(self.builder, item)
        else:
            raise ValueError(f'Could not find {item}')

    def with_initial_contigs(self, *contigs):
        self.initial_contigs = contigs
        return self

    def run(self):
        shutil.rmtree(self.tmpdir)
        self.tmpdir.mkdir()
        graph = self.builder.build(self.tmpdir)
        out_dir = self.tmpdir / 'abeona_output'
        command = ['subgraphs', graph, out_dir]

        if self.initial_contigs:
            initial_contigs = self.tmpdir / 'initial-kmers.fa'
            records = [
                SeqRecord(Seq(contig), id=f'initial_contig_{idx}') for idx, contig in
                enumerate(self.initial_contigs)
            ]
            SeqIO.write(records, str(initial_contigs), 'fasta')
            command += ['--initial-contigs', initial_contigs]

        args = ['abeona'] + [str(arg) for arg in command]
        abeona_main(args)

        graphs = list(Path(out_dir).glob('g*.traverse.ctx'))

        return Traversals(graphs)


@attr.s(slots=True)
class ReadsTestDriver:
    builder = attr.ib(attr.Factory(Mccortex))
    record_buffer_size = attr.ib(None)

    def __getattr__(self, item):
        if item in ['with_kmer_size', 'with_dna_sequence', 'with_dna_sequence_pair', 'with_min_unitig_coverage']:
            return getattr(self.builder, item)
        else:
            raise ValueError(f'Could not find {item}')

    def with_record_buffer_size(self, n):
        self.record_buffer_size = n
        return self

    def run(self, tmpdir):
        out_dir = tmpdir / 'abeona_reads'
        traversal_expectation = SubgraphTestDriver(tmpdir, self.builder).run()
        graphs = traversal_expectation.traversals

        for idx in [1, 2]:
            subprocess.run(
                f'cat {tmpdir}/*.{idx}.fasta > {tmpdir}/combined.{idx}.fasta',
                shell=True,
                check=True
            )

        graph_list = tmpdir / 'subgraph_list.txt'
        with open(graph_list, 'w') as fh:
            fh.write('# prefix\tgraph\n')
            for graph in graphs:
                fh.write(f'{out_dir/graph.stem}\t{graph}\n')

        command = f'abeona reads --format fasta {graph_list} {tmpdir}/combined.1.fasta'
        if self.builder.is_paired:
            command += f' --reverse {tmpdir}/combined.2.fasta'
        if self.record_buffer_size:
            command += f' --record-buffer-size {self.record_buffer_size}'
        abeona_main(command.split())

        reads = list(Path(out_dir).glob('g*.fa.gz'))

        return TraversalsAndFastas(Fastas(reads), traversal_expectation)



import gzip
from logging import getLogger
from pathlib import Path

import attr
from cortexpy.graph.parser.random_access import RandomAccess

from .utils import get_maybe_gzipped_file_handle, get_fastx_title_seq_generator

logger = getLogger('abeona.reads')


@attr.s(slots=True)
class GraphData:
    prefix = attr.ib()
    graph_path = attr.ib()
    format = attr.ib()
    is_paired = attr.ib()
    buffer1 = attr.ib(attr.Factory(list))
    buffer2 = attr.ib(attr.Factory(list))
    _fh = attr.ib(None)
    _file1 = attr.ib(init=False)
    _file2 = attr.ib(init=False)

    def __attrs_post_init__(self):
        Path(self.prefix).parent.mkdir(exist_ok=True)
        self._file1 = self.prefix + f'.1.fa.gz'
        path = Path(self._file1)
        if path.is_file():
            path.unlink()
        if self.is_paired:
            self._file2 = self.prefix + f'.2.fa.gz'
            path = Path(self._file2)
            if path.is_file():
                path.unlink()

    @property
    def fh(self):
        if self._fh is None:
            self._fh = [gzip.open(self._file1, 'at')]
            if self.is_paired:
                self._fh.append(gzip.open(self._file2, 'at'))
        return self._fh

    def store_record_pair(self, *records):
        self.buffer1.append(records[0])
        if self.is_paired:
            self.buffer2.append(records[1])

    def write_record_pair(self, *records):
        """Accepts one or two reads. Figures out if one or two reads are expected."""
        self.fh[0].write(''.join(['>', records[0][0], '\n', records[0][1], '\n']))
        if self.is_paired:
            self.fh[1].write(''.join(['>', records[1][0], '\n', records[1][1], '\n']))

    def flush(self):
        buffers = [self.buffer1]
        if self.is_paired:
            buffers.append(self.buffer2)
        for recs in zip(*buffers):
            self.write_record_pair(*recs)
        self.fh[0].close()
        self.buffer1.clear()
        if self.is_paired:
            self.fh[1].close()
            self.buffer2.clear()
        self._fh = None


def main(args):
    record_buffer_size = args.record_buffer_size

    is_paired = args.reverse is not None
    graphs = []
    with open(args.graph_list) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            prefix, graph = line.rstrip().split('\t')
            graphs.append(GraphData(prefix=prefix,
                                    graph_path=graph,
                                    format=args.format,
                                    is_paired=is_paired))
    logger.info(f'Assigning reads to {len(graphs)} graphs')

    kmers = {}
    kmer_size = None
    for idx, graph_data in enumerate(graphs):
        with open(graph_data.graph_path, 'rb') as fh:
            ra = RandomAccess(fh)
            if kmer_size is None:
                kmer_size = ra.kmer_size
                logger.info(f'Detected kmer size: {kmer_size}')
            assert kmer_size == ra.kmer_size
            for kmer in ra:
                assert kmer not in kmers, 'Input subgraphs must be disjoint'
                kmers[kmer] = idx

    assert kmer_size, "Could not find any graphs"

    reads = [get_fastx_title_seq_generator(get_maybe_gzipped_file_handle(args.forward, 'rt'))]
    if is_paired:
        reads.append(
            get_fastx_title_seq_generator(get_maybe_gzipped_file_handle(args.reverse, 'rt')))
    n_records_stored = 0
    for recs in zip(*reads):
        seen_graphs = set()
        for read in (s for t, s in recs):
            for start in range(len(read) - kmer_size + 1):
                kmer = read[start:(start + kmer_size)]
                if kmer in kmers:
                    if record_buffer_size is not None and record_buffer_size >= n_records_stored:
                        for graph in graphs:
                            graph.flush()
                        n_records_stored = 0
                    idx = kmers[kmer]
                    if idx not in seen_graphs:
                        graphs[idx].store_record_pair(*recs)
                        n_records_stored += 1
                        seen_graphs.add(idx)
    for graph in graphs:
        graph.flush()



def is_gz_file(filepath):
    import binascii
    with open(filepath, 'rb') as test_f:
        return binascii.hexlify(test_f.read(2)) == b'1f8b'


def get_maybe_gzipped_file_handle(*args):
    import gzip
    if is_gz_file(args[0]):
        return gzip.open(*args)
    else:
        return open(*args)



import json
from pathlib import Path
from contextlib import closing

import attr
from cortexpy.graph.parser.streaming import load_cortex_graph
from cortexpy.test.expectation import KmerGraphExpectation, Fasta
from .utils import get_maybe_gzipped_file_handle

@attr.s(slots=True)
class AbeonaKmerGraphExpectation(KmerGraphExpectation):
    graph_path = attr.ib(None)
    meta = attr.ib(None)

    def has_meta_info(self, key, val):
        assert val == self.meta[key]
        return self


@attr.s(slots=True)
class Traversals:
    traversals = attr.ib()
    traversal_expectations = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.traversal_expectations = [
            AbeonaKmerGraphExpectation(
                load_cortex_graph(open(graph, 'rb')),
                graph_path=graph,
                meta=json.load(open(str(graph) + '.json', 'r'))
            ) for graph in self.traversals
        ]

    def has_graph_with_kmers(self, *kmers):
        kmers = set(kmers)
        tested_kmers = []
        for expect in self.traversal_expectations:
            tested_kmers.append(sorted(expect.graph))
            if sorted(kmers) == tested_kmers[-1]:
                return expect
        assert False, f'Could not find graph with kmers: {kmers}\nin: {tested_kmers}'

    def has_n_graphs(self, n):
        assert n == len(self.traversal_expectations)
        return self


@attr.s(slots=True)
class Fastas:
    fastas = attr.ib()

    def has_fastas_with_prefix(self, stem):
        fastas = []
        for fasta in sorted(self.fastas):
            if Path(fasta).with_suffix('').stem.startswith(stem):
                with closing(get_maybe_gzipped_file_handle(fasta, 'rt')) as fh:
                    fastas.append(Fasta(fh.read()))
        if len(fastas) == 0:
            assert False, f'Could not find stem ({stem}) in fastas\n{self.fastas}'
        return fastas


@attr.s(slots=True)
class TraversalsAndFastas:
    fastas_expection = attr.ib()
    traversals_expectation = attr.ib()

    def has_graph_with_kmers_and_mapped_reads(self, kmers, seqs, seqs2=None):
        graph = self.traversals_expectation.has_graph_with_kmers(*kmers).graph_path
        graph_stem = Path(graph).stem

        fasta_expects = self.fastas_expection.has_fastas_with_prefix(graph_stem)

        seq_lists = [seqs]
        if seqs2 is not None:
            seq_lists.append(seqs2)
        assert len(fasta_expects) == len(seq_lists), f'len({fasta_expects}) != len({seq_lists})'
        for idx, seq_list in enumerate(seq_lists):
            for seq in seq_list:
                fasta_expects[idx].has_record(seq)
            fasta_expects[idx].has_n_records(len(seq_list))
        return self

    def has_n_graphs(self, n):
        self.traversals_expectation.has_n_graphs(n)
        return self



import logging
import sys

logger = logging.getLogger('abeona')


def main(argv=sys.argv):
    from . import __version__
    import argparse
    subcommands = {
        'assemble': assemble_main,
        'subgraphs': subgraphs_main,
        'reads': reads_main,
    }
    parser = argparse.ArgumentParser(description=f'abeona version {__version__}', prog='abeona')
    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s version {__version__}')
    parser.add_argument('sub_command', choices=sorted(subcommands.keys()),
                        help='abeona sub-command')
    parser.add_argument('args', nargs=argparse.REMAINDER, help='sub-command arguments')
    args = parser.parse_args(argv[1:])

    import warnings
    from Bio import BiopythonWarning
    warnings.simplefilter('ignore', BiopythonWarning)
    return subcommands[args.sub_command](args.args)


def assemble_main(argv):
    import argparse
    parser = argparse.ArgumentParser(description='Run abeona assembly pipeline.',
                                     prog='abeona assemble',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-o', '--out-dir', help='Output directory', required=True)
    parser.add_argument('-j', '--jobs', help='Number of jobs to schedule concurrently', default=2,
                        type=int)
    parser.add_argument('-k', '--kmer-size', default=47,
                        help='k-mer size to use to construct the De Bruijn graph')
    parser.add_argument('-m', '--memory', default=3, help='Maximum memory to use in giga bytes',
                        type=int)
    parser.add_argument('-q', '--quiet', action='store_true')
    parser.add_argument('--resume', action='store_true')
    parser.add_argument('--no-cleanup', action='store_true')
    parser.add_argument('--with-report', action='store_true',
                        help='Create nextflow report file at nexttflow_report.html in output directory')
    parser.add_argument('--with-dag', action='store_true',
                        help='Create flowchart of workflow at flowchart.png in output directory')

    parser.add_argument('--fastx-forward', help='Forward sequences in FASTA/FASTQ format',
                        required=False)
    parser.add_argument('--fastx-reverse', help='Reverse sequences in FASTA/FASTQ format',
                        required=False)
    parser.add_argument('--fastx-single', help='Single-end sequences in FASTA/FASTQ format',
                        required=False)
    parser.add_argument('--initial-contigs', help='Only start assembly from contigs in this FASTA',
                        required=False)
    parser.add_argument('--extra-start-kmer',
                        help='Disconnect this k-mer from incoming k-mers before '
                             'candidate transcript creation. '
                             'This may be useful when assembling circular genomes. '
                             'Best used with --initial-contigs to make sure the k-mer exists '
                             'in the consistent cortexpy graph.',
                        required=False)

    group = parser.add_argument_group('Graph traversal cleaning')
    group.add_argument('--min-tip-length', type=int, default=-1,
                       help='Prune tips shorter than this value.'
                            ' A value of -1 sets the min tip length to the value of --kmer-size')
    group.add_argument('--min-unitig-coverage', type=int, default=4,
                       help="Prune unitigs with mean coverage below this value")
    group.add_argument('--no-prune-tips-with-mccortex', action='store_true',
                       help="Instead of Mccortex use cortexpy to prune unitigs. "
                            "This is slower and not recommended at this time.")
    group.add_argument('--prune-tips-with-mccortex', action='store_true')
    group.add_argument('--prune-tips-iteratively', action='store_true',
                       help="Prune the graph of tip lengths x = 2^n while x is less than --min-tip-length. "
                            "Finally, prune graph of tips shorter than --min-tip-length. "
                            "Currently only works when pruning with Mccortex. ")

    group = parser.add_argument_group('Candidate transcript creation')
    group.add_argument('--max-paths-per-subgraph', type=int, default=0,
                       help='Ignore graphs that have more than this number of paths')
    group.add_argument('--no-links', action='store_true',
                       help='Do not use links in candidate transcript creation')
    group.add_argument('--report-unassembled-reads', action='store_true',
                       help='Store reads from ignored graphs in unassembled_reads')
    group.add_argument('--assemble-unassembled-reads-with-transabyss', action='store_true',
                       help='Try and assemble reads from ignored graphs with transabyss')

    group = parser.add_argument_group('kallisto arguments',
                                      description='Arguments passed directly on to kallisto')
    group.add_argument('--bootstrap-samples', type=int, default=100,
                       help='Number of kallisto bootstrap samples')
    group.add_argument('--kallisto-fragment-length', type=float)
    group.add_argument('--kallisto-sd', type=float)
    group.add_argument('--kallisto-threads', type=int, default=2,
                       help='Number of logical cores to assign to a single kallisto quant job. '
                            'Needs to be less than or equal to --jobs')
    group.add_argument('--max-read-length', type=int,
                       help='Length of longest read in data. Remove candidate subgraphs that do '
                            'not have at least one candidate transcript greater than this length. '
                            'Kallisto cannot align reads to transcripts that are shorter than the '
                            'read. If this value is not specified, then abeona estimates this '
                            'value from the head of the reads supplied to kallisto '
                            '(--kallisto-fastx-*).')

    group = parser.add_argument_group('Candidate transcript filtering')
    group.add_argument('--estimated-count-threshold', type=float, default=1,
                       help='Threshold over which the estimated transcript count from kallisto is '
                            'counted towards keeping a transcript')
    group.add_argument('--bootstrap-proportion-threshold', type=float, default=0.95,
                       help="Proportion of bootstrap iterations for which a transcript's estimated "
                            "counts must be above the --estimated-count-threshold")
    group.add_argument('--record-buffer-size', type=int, default=-1,
                       help='Number of reads to buffer in memory when assigning reads to subgraphs')
    group.add_argument('--max-junctions', type=int, default=0,
                       help='The max junctions argument can be used to quickly ignore large subgraphs'
                            ' with too many junctions to process effectively.')

    args = parser.parse_args(args=argv)
    args = preprocess_assemble_args(args)
    validate_assemble_args(args, parser)

    from pathlib import Path
    import subprocess

    out_dir = Path(args.out_dir)
    out_dir.mkdir(exist_ok=True)

    script_name = 'assemble.nf'
    pipeline_script = out_dir / script_name
    package_script = Path(__file__).parent / 'assemble' / script_name
    if pipeline_script.is_symlink():
        pipeline_script.unlink()
    pipeline_script.symlink_to(package_script)

    args_file = write_args(args, out_dir)

    cmd = f'cd {out_dir} && KALLISTO_THREADS={args.kallisto_threads} JOBS={args.jobs} nextflow run {script_name} -process.maxForks {args.jobs} -params-file {args_file}'
    if args.with_report:
        cmd += ' -with-report nexttflow_report.html'
    if args.with_dag:
        cmd += ' -with-dag flowchart.png'
    if args.resume:
        cmd += ' -resume'
    logger.info(cmd)
    cprocess = subprocess.run(cmd, shell=True)
    if cprocess.returncode != 0:
        return cprocess.returncode
    gzipped_transcripts = out_dir / "all_transcripts" / "transcripts.fa.gz"
    unzipped_transcripts = out_dir / "transcripts.fa"
    if gzipped_transcripts.is_file():
        subprocess.run(
            f'gzip -dc {gzipped_transcripts} > {unzipped_transcripts}',
            shell=True
        )
    else:
        unzipped_transcripts.touch(exist_ok=True)

    import shutil
    if not args.no_cleanup:
        for d in out_dir.iterdir():
            if d.is_dir() and not d.name.startswith('.') and not d.name == 'unassembled_reads':
                shutil.rmtree(d)
    return cprocess.returncode


def write_args(args, out_dir):
    import json

    args_file = 'args.json'
    args_dict = {a: getattr(args, a) for a in dir(args) if not a.startswith('_')}
    args_dict['mccortex'] = f'mccortex {args.kmer_size}'
    args_dict['mccortex_args'] = f'--sort --force -m {args.memory}G'
    if args.max_read_length is None:
        max_read_length = estimate_max_read_length(args)
        assert max_read_length is not None
        args_dict['max_read_length'] = max_read_length
    args_dict['mccortex_thread_args'] = f'--force -m {args.memory // args.jobs}G'
    with open(out_dir / args_file, 'w') as fh:
        json.dump(args_dict, fh)
    return args_file


def estimate_max_read_length(args):
    from .utils import get_maybe_gzipped_file_handle, get_fastx_title_seq_generator
    from itertools import islice
    n_reads_to_read = 100
    max_read_length = None
    for arg in ['fastx_single', 'fastx_forward', 'fastx_reverse']:
        file = getattr(args, arg)
        if file is not None:
            with get_maybe_gzipped_file_handle(file, 'rt') as fh:
                for _, seq in islice(get_fastx_title_seq_generator(fh), n_reads_to_read):
                    if max_read_length is None or len(seq) > max_read_length:
                        max_read_length = len(seq)
    return max_read_length


def subgraphs_main(argv):
    import argparse
    parser = argparse.ArgumentParser(description='Partition cortex graph into its subgraphs.',
                                     prog='abeona subgraphs')
    parser.add_argument('graph')
    parser.add_argument('out_dir')
    parser.add_argument('-m', '--memory', type=int, default=3)
    parser.add_argument('-c', '--cores', type=int, default=2)
    parser.add_argument('--initial-contigs', help='Only start assembly from contigs in this FASTA',
                        required=False)

    args = parser.parse_args(argv)

    from .subgraphs import main
    return main(args)


def reads_main(argv):
    import argparse
    parser = argparse.ArgumentParser(description='Assign reads to cortex graphs.',
                                     prog='abeona reads')
    parser.add_argument('graph_list',
                        help="""
                        Tab delimited text file with two columns and optional header line.
                        All lines starting with '#' are ignored. For example:
                        ```
                        # prefix\tgraph
                        out_dir/g1\tg1.ctx
                        out_dir/g2\tg2.ctx
                        ```
                        """)
    parser.add_argument('forward', help='Forward reads file in FASTA or FASTQ format.')
    parser.add_argument('--reverse',
                        help='Reverse reads file in FASTA or FASTQ format. '
                             'Only specified if reads are paired',
                        default=None)
    parser.add_argument('--format', help='File format of input reads.',
                        choices=['fasta', 'fastq'],
                        required=True)
    parser.add_argument('--record-buffer-size', default=None, type=int,
                        help='Flush all reads to disk after this many records have been assigned')

    args = parser.parse_args(argv)

    from .reads import main
    return main(args)


def preprocess_assemble_args(args):
    if args.min_tip_length == -1:
        args.min_tip_length = args.kmer_size
    if not args.prune_tips_with_mccortex and not args.no_prune_tips_with_mccortex:
        args.prune_tips_with_mccortex = True
    make_file_paths_absolute(args)

    return args


def validate_assemble_args(args, parser):
    if (args.fastx_forward is None) != (args.fastx_reverse is None):
        raise parser.error(
            'Both --fastx-forward and --fastx-reverse must be specified or not specified')
    if (args.fastx_forward is None) == (args.fastx_single is None):
        raise parser.error('Need to specify --fastx-forward or --fastx-single')
    if args.fastx_single is not None:
        if args.kallisto_fragment_length is None:
            raise parser.error('Required: --kallisto-fragment-length')
        if args.kallisto_sd is None:
            raise parser.error('Required: --kallisto-sd')
    if int(args.jobs) < int(args.kallisto_threads):
        raise parser.error('--jobs needs to be greater or equal to --kallisto-threads')
    assert not (args.prune_tips_with_mccortex and args.no_prune_tips_with_mccortex)
    if args.prune_tips_iteratively:
        assert args.prune_tips_with_mccortex


def make_file_paths_absolute(args):
    from pathlib import Path
    for file_arg in [
        'fastx_forward', 'fastx_reverse', 'fastx_single', 'initial_contigs',
    ]:
        file = getattr(args, file_arg)
        if file is not None:
            file = Path(file)
            assert file.is_file(), f'Not a file: {file}'
            file = file.absolute()
            assert file.is_file(), f'Not a file: {file}'
            setattr(args, file_arg, str(file))



import json
import logging
from pathlib import Path

import attr
import progressbar
from Bio import SeqIO
from cortexpy.constants import EngineTraversalOrientation
from cortexpy.graph.parser.random_access import SlurpedRandomAccess
from cortexpy.graph.serializer.kmer import dump_colored_de_bruijn_graph_to_cortex
from cortexpy.graph.traversal.engine import Engine
from cortexpy.utils import lexlo

progressbar.streams.wrap_stderr()
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('abeona.subgraphs')


def load_initial_kmers(random_access, initial_contigs):
    logger.info('Subsetting graph kmer strings on initial-contig kmers')
    initial_kmers = set()
    first_kmer = next(iter(random_access))
    for rec in SeqIO.parse(initial_contigs, 'fasta'):
        kmer_size = len(first_kmer)
        assert len(rec.seq) >= kmer_size
        for start_idx in range(len(rec.seq) - kmer_size + 1):
            initial_kmers.add(str(lexlo(rec.seq[start_idx:start_idx + kmer_size])))
    logger.info(f'Loaded {len(initial_kmers)} initial kmers')
    k_string_tracker = KmerStringTracker(
        unseen=sorted(kmer for kmer in initial_kmers if kmer in random_access))
    logger.info(
        f'Keeping {len(k_string_tracker.unseen)} out of {len(initial_kmers)} initial kmers')
    return k_string_tracker


@attr.s(slots=True)
class KmerStringTracker(object):
    unseen = attr.ib(attr.Factory(list))
    seen = attr.ib(attr.Factory(set))

    def add_seen(self, string):
        self.seen.add(string)

    def strings_not_seen(self):
        for string in self.unseen:
            if string not in self.seen:
                yield string


def count_junctions(graph):
    """Return the number of junctions in graph"""
    junc_count = 0
    for node in graph.nodes():
        if graph.in_degree(node) > 1:
            junc_count += 1
        if graph.out_degree(node) > 1:
            junc_count += 1
    return junc_count


def main(args):
    out_dir = Path(args.out_dir)
    out_dir.mkdir(exist_ok=True)

    input_graph = Path(args.graph)
    if not input_graph.is_file():
        raise Exception(f'Input cortex graph ({input_graph}) does not exist')

    logger.info(f'Slurping cortex graph: {input_graph}')
    ra = SlurpedRandomAccess.from_handle(open(input_graph, 'rb'))
    if args.initial_contigs:
        kstring_tracker = load_initial_kmers(ra, args.initial_contigs)
    else:
        kstring_tracker = KmerStringTracker(unseen=ra)
    graph_idx = 0

    with progressbar.ProgressBar(max_value=len(ra)) as bar:
        for initial_kmer_string in kstring_tracker.strings_not_seen():
            graph_id = f'g{graph_idx}'
            subgraph_path = out_dir / f'{graph_id}.traverse.ctx'

            logger.debug('Traversing graph %s', graph_idx)
            engine = Engine(
                ra,
                orientation=EngineTraversalOrientation.both,
                max_nodes=None,
                logging_interval=90
            )
            engine.traverse_from(str(initial_kmer_string))

            logger.debug('Writing graph %s to: %s', graph_idx, subgraph_path)
            with open(subgraph_path, 'wb') as out_fh:
                dump_colored_de_bruijn_graph_to_cortex(engine.graph, out_fh)
            with open(str(subgraph_path) + '.json', 'wt') as json_fh:
                json.dump({'n_junctions': count_junctions(engine.graph)}, json_fh)
            for node in engine.graph:
                kstring_tracker.add_seen(lexlo(node))
            logger.debug(
                'Found subgraph with %s kmers - at most %s kmers left', len(engine.graph),
                len(ra) - len(kstring_tracker.seen))
            bar.update(len(kstring_tracker.seen))
            graph_idx += 1
    logger.info('No kmers remaining')



def is_gz_file(filepath):
    import binascii
    with open(filepath, 'rb') as test_f:
        return binascii.hexlify(test_f.read(2)) == b'1f8b'


def get_maybe_gzipped_file_handle(*args):
    import gzip
    if is_gz_file(args[0]):
        return gzip.open(*args)
    else:
        return open(*args)


def get_fastx_title_seq_generator(fh):
    """Requires a seekable file handle. Needs file to start at beginning of file handle."""
    from Bio.SeqIO.QualityIO import FastqGeneralIterator
    from Bio.SeqIO.FastaIO import SimpleFastaParser

    # determine if the file looks like a fastq
    fh.seek(0)
    is_fastq = True
    try:
        next(FastqGeneralIterator(fh), None)
    except ValueError as e:
        if e.args[0] == "Records in Fastq files should start with '@' character":
            is_fastq = False
        else:
            raise
    fh.seek(0)
    if is_fastq:
        return ((title, seq) for title, seq, qual in FastqGeneralIterator(fh))
    else:
        return SimpleFastaParser(fh)
